"""
This module provides an interface to SMV files which often end in the IMG format.
These are commonly used in microED.

See https://wiki.uni-konstanz.de/ccp4/index.php/SMV_file_format for more details.

Note
----
General users:
    Use the simplified smv.smvReader() function to load the data and meta
    data as a python dictionary.

Advanced users and developers:
    Access the file internals through the smv.fileSMV() class.

"""

from pathlib import Path
import datetime

import numpy as np

class fileSMV:
    """Class to represent SMV files.

    Attributes
    ----------
    file_name : str
        The name of the file
    file_path : pathlib.Path
        A pathlib.Path object for the open file
    fid : file
        The file handle to the opened MRC file.
    dataType : np.dtype
        The numpy dtype of the data.
    dataSize : np.ndarray
        The number of pixels along each dimension. Corresponds to the shape attribute of a np.ndarray
    num_header_bytes : 
        The number of bytes in the header. Usually 512.
    header_info : dict
        A dictionary containing the header meta data.

    camera_length=110, lamda=0.0197, pixel_size=0.01, beam_center=None, binned_by=1
    """
    
    def __init__(self, filename, verbose=False):
        """ Initialize opening the file
        
        Parameters
        ----------
        filename : str or pathlib.Path or file object
            String pointing to the filesystem location of the file.

        verbose : bool, optional, default False
            If True, debug information is printed.
            
        """
        
        self._verbose = verbose
        self._expected_keys = ('HEADER_BYTES', 'DIM', 'BYTE_ORDER', 'TYPE', 'SIZE1', 'SIZE2', 'PIXEL_SIZE', 
                         'WAVELENGTH', 'DISTANCE', 'PHI', 'BEAM_CENTER_X', 'BEAM_CENTER_Y', 'BIN', 
                         'DATE', 'DETECTOR_SN', 'OSC_RANGE', 'OSC_START', 'IMAGE_PEDESTAL', 'TIME', 
                         'TWOTHETA')
        self._data_types = {'unsigned_short': np.uint16} # convert SMV types with numpy dtypes
        self.header_info = {}
        self.num_header_bytes = None
        self.dataType = None
        self.dataSize = [0, 0]
        self._v = verbose
        
        if hasattr(filename, 'read'):
            self.fid = filename
            try:
                self.file_name = self.fid.name
            except AttributeError:
                self.file_name = None
        else:
            # check filename type. Prefer pathlib.Path
            if isinstance(filename, str):
                filename = Path(filename)
            elif isinstance(filename, Path):
                pass
            else:
                raise TypeError('Filename is supposed to be a string or pathlib.Path')
        
        self.file_path = filename
        self.file_name = self.file_path.name
        
        try:
            self.fid = open(self.file_path, 'rb')
        except IOError:
            print('Error reading file: "{}"'.format(self.file_path))
            raise
        except:
            raise
        
        if not self._validate():
            raise IOError('Not an SMV file: {}'.format(self.file_path))
        
        self.readHeader()
        self.parseHeader()
        
    def __del__(self):
        """Destructor which also closes the file

        """
        if not self.fid.closed:
            if self._v:
                print('Closing input file: {}'.format(self.file_path))
            self.fid.close()

    def __enter__(self):
        """Implement python's with statement

        """
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        """Implement python's with statement
        and close the file via __del__()
        """
        self.__del__()
        return None
    
    def _validate(self):
        # first_line = self.fid.read(15).decode('UTF-8')
        first_line = self.fid.readline() # should be {
        second_line = self.fid.readline().decode('UTF-8') # should contain HEADER_BYTES
        if 'HEADER_BYTES=' in second_line:
            self.num_header_bytes = int(second_line.split('=')[1].strip().strip(';'))
            return True
        else:
            return False
    
    def readHeader(self):
        """Read the header information and conver to numbers or strings."""
        
        self.fid.seek(0, 0)
        head = self.fid.read(self.num_header_bytes).decode('UTF-8').splitlines()
        for line in head:
            if '=' in line:
                key, val = line.split('=')
                val = val.strip(';')
                if key in self._expected_keys:
                    try:
                        self.header_info[key] = float(val)
                        try:
                            self.header_info[key] = int(val)
                        except:
                            pass # it is a float
                    except:
                        self.header_info[key] = val # not a number
    
    def parseHeader(self):
        """Parse the header dictionary for relelvant information to read the data in the file."""
        for key, val in self.header_info.items():
            if key == 'SIZE1':
                self.dataSize[1] = val # column
            elif key == 'SIZE2':
                self.dataSize[0] = val # row
            elif key == 'TYPE':
                try:
                    self.dataType = self._data_types[val]
                except KeyError:
                    raise(f'File data type not supported: {val}')
    
    def getDataset(self):
        self.readHeader()
        self.parseHeader()
        
        self.fid.seek(self.num_header_bytes, 0)
        data = np.fromfile(self.fid, count=self.dataSize[0] * self.dataSize[1], dtype=self.dataType)
        data = data.reshape(self.dataSize)
        data_out = {}
        data_out['data'] = data
        return data_out
    
def smvWriter(out_path, dp, camera_length=110, lamda=0.0197, pixel_size=0.01, beam_center=None, binned_by=1, newline=None):
    """ Write out data as a SMV (.img) formatted file
    Header is 512 bytes of zeros and then filled with ASCII.
    
    Note:
    - only little endian is supported
    - only uint16 is supported
    - ony 2D data is supported
    - some other meta data (PHI, DATE, etc.) is populated with hard coded values
    
    Parameters
    ----------
    camera_length : float
        The calibrated camera length (not the label) in mm. Default is 110 mm.
    lamda : float
        The wavelength of the radiation in Ansgroms. Default is 0.0197 for 300 kV electrons
    pixel_size : float
        Physical detector pixel size in mm. Default is 0.01 mm (10 microns)
    beam_center : tuple
        The location of the center beam in column, row format in mm (not pixels!)
    binned_by : int
        The binning applied to the original data. This is necessary for proper
        calibrations of detector distances and beam center. Default is 1.
    newline : str (optional)
        Allow the user to specify the newline character. For data written on Windows computers
        some processing programs in Linux are not able to load SMV files with Windows 
        carriage return and newline characters. Use '\n' on Windows machines to enforce Linux
        line endings. The default `None` will use the system default.
    """
    if dp.dtype != np.uint16:
        raise TypeError("Only uint16 data type is supported.")
    dtype = 'unsigned_short'
    
    if not beam_center:
        beam_center = [ii / 2 * pixel_size for ii in dp.shape]
    
    # make sure binned_by is an integer
    binned_by = int(binned_by)
    
    # Write 512 bytes of zeros
    with open(out_path, 'wb') as f0:
        f0.write(np.zeros(512, dtype=np.uint8))
    # Write the header over the zeros as needed
    # The newline character is system default unless otherwise specified
    with open(out_path, 'r+', newline=newline) as f0:
        f0.write("{\nHEADER_BYTES=512;\n")
        f0.write("DIM=2;\n")
        f0.write("BYTE_ORDER=little_endian;\n")
        f0.write(f"TYPE={dtype};\n")
        f0.write(f"SIZE1={dp.shape[1]};\n")  # size 1 is columns
        f0.write(f"SIZE2={dp.shape[0]};\n")  # size 2 is rows
        f0.write(f"PIXEL_SIZE={pixel_size};\n")  # physical pixel size in mm
        f0.write(f"WAVELENGTH={lamda};\n")  # wavelength in Angstroms
        f0.write(f"DISTANCE={int(camera_length)};\n") # in mm
        f0.write("PHI=0.0;\n")
        f0.write(f"BEAM_CENTER_X={beam_center[1]};\n") # in mm (not pixels!)
        f0.write(f"BEAM_CENTER_Y={beam_center[0]};\n") 
        f0.write(f"BIN={binned_by}x{binned_by};\n")
        f0.write("DATE=Fri Dec 31 23:59:59 1999;\n")
        f0.write("DETECTOR_SN=unknown;\n")
        f0.write("OSC_RANGE=1.0;\n")
        f0.write("OSC_START=0;\n")
        f0.write("IMAGE_PEDESTAL=0;\n")
        f0.write("TIME=1.0;\n")
        f0.write("TWOTHETA=0;\n")
        f0.write("}\n")
    # Append the binary image data at the end of the header
    with open(out_path, 'rb+') as f0:
        f0.seek(512, 0)
        f0.write(dp)

        
def smvReader(file_name, verbose=False):
    """ A simple function to read open a SMV, parse the header, and read the
    data and meta data.

    Parameters
    ----------
        file_name : str or pathlib.Path
            The path to the file to load.

    Returns
    -------
        out : dict
            A dictionary containing the data and interesting metadata.
    
    Note
    ----
    The returned dictionary has an entry called `pixelSize`. This is the calibrated distance
    in angstroms. It is not the physical size of a detector pixel. 

    Example
    -------
        Simply read in all data from disk into memory. This assumes the dataset is 3 dimensional:
        >> from ncempy.io.smv import smvReader
        >> import matplotlib.pyplot as plt
        >> mrc1 = smvReader.('filename.mrc')
        >> plt.imshow(mrc1['data'][0, :, :]) #show the first image in the data set
    """
    if isinstance(file_name, str):
        file_name = Path(file_name)

    with fileSMV(file_name) as f1:  # open the file and init the class
        im1 = f1.getDataset()  # read in the dataset

        # Calculate the pixel size in inverse angstroms according to the geometry in the header
        BIN = [int(ii) for ii in f1.header_info['BIN'].split('x')]
        alpha = (BIN[0] * f1.header_info['PIXEL_SIZE']) / f1.header_info['DISTANCE'] # angle across 1 pixel
        dp_pixel_distance = alpha / f1.header_info['WAVELENGTH'] * 1e-10 # divide by wavelength to get distance in Angstroms
        pixelSize = (dp_pixel_distance, dp_pixel_distance)
        extra_metadata = {'pixelSize': pixelSize, 'pixelUnit':'A', 'filename': f1.file_name}
    im1.update(extra_metadata)
    return im1




import numpy as np
from scipy import ndimage

from . import radial_profile
from . import peak_find

def rebin(im, f, funcType='sum'):
    """ Rebin a 2D array. From stackoverflow:
    https://stackoverflow.com/questions/4624112/grouping-2d-numpy-array-in-average

    This should be used carefully. There is not antialiasing applied which could
    produce odd effects for some data sets (such as high resolution data)

    Parameters
    ----------
    im : ndarray
        2D array to reduce
    f : int
        The factor to rebin by. Must be integer
    funcType : str
        The type of reduction. mean or sum are implemented.

    Returns
    -------
    : ndarray
        The 2D array with the new size.
    """

    nbig = im.shape
    nsmall = [ii // f for ii in im.shape]

    # Reshape the array so that the required neighborhood are in arrays along certain axes
    # Then average or sum those neighborhoods
    im_reshape = im.reshape([nsmall[0], nbig[0] // nsmall[0], nsmall[1], nbig[1] // nsmall[1]])
    # Reduce using different types of functions
    if funcType == 'mean':
        small = im_reshape.mean(3).mean(1)
    elif funcType == 'sum':
        small = im_reshape.sum(3).sum(1)
    # elif funcType == 'median':
    #    small = im_reshape.median(3).median(1)
    else:
        print('defaulting to sum')
        small = im_reshape.sum(3).sum(1)

    return small

##
# Image alignment codes
##

def image_cross_corr(image, reference, real_filter=1, k_filter=1):
    """ Calculate image cross-correlation. See imageCrossCorRealShift and other
    similar functions to calculate the shift and apply the shift.

    Parameters
    ----------
        image : ndarray
            The source image to align to. Should be even sized.

        reference : ndarray
            The reference image to align to the source image. Should be even sized.

        real_filter : ndarray, optional, default 1
            Filter to apply to each image in real space before FFT.

        k_filter : ndarray, optional default 1
            Filter to apply to each image in FFT space

    Returns
    -------
        : ndarray
            Cross correlation of image and reference.

    """

    if type(image) is not np.ndarray or type(reference) is not np.ndarray:
        raise TypeError("Must use ndarrays")

    if np.iscomplexobj(image) or np.iscomplexobj(reference):
        raise TypeError("Images mst be real")

    # Calculate FFT of images
    image_f = np.fft.fft2((image - np.mean(image)) * real_filter)
    reference_f = np.fft.fft2((reference - np.mean(reference)) * real_filter)

    # Calculate cross-correlation
    xcor = abs(np.fft.fftshift(np.fft.ifft2(np.conj(image_f) * reference_f * k_filter)))

    return xcor


def image_correlate(image, reference, real_filter=1, k_filter=1, shift_func='shift', verbose=False):
    """ Align image to reference by cross-correlation. Outputs shifts and shifted images.
    Uses the real FFT for ~2x speed improvement. The k_filter must have
    a shape that matches the np.fft.rfft2() of image and reference.
    Uses scipy.ndimage.shift() or np.roll to move the image. Use 'roll' to avoid losing
    data off the edge for multiple shifting operations. Use shift to avoid wrap around problems and when there
    is only one shifting operation.

    Note
    ----
        image, reference and real_filter must all have the same shape (N, M).
        k_filter must have a shape that matches the np.fft.rfft2() of
        the other inputs: (N, M/2+1)

    Parameters
    ----------
        image : ndarray
            A image as a 2D ndarray.

        reference : ndarray
            The reference image to align to.

        real_filter : ndarray, optional, default = 1
            A real space filter applied to image and reference before
            calculating the shift.

        k_filter : ndarray, optional, default = 1
            A Fourier space filter applied to the fourier transform of
            image and reference before calculating the cross-correlation.

        shift_func : str, default is 'shift'
            The function to use to shift the images. 'roll' uses np.roll and 'shift' uses ndimage.shift.

        verbose : bool
            Plots the cross-correlation using matplotlib for debugging purposes.

    Returns
    -------
        : tuple, (ndarray, tuple)
            A tuple containing the shifted image and the shifts applied.
    """
    output = None

    if shift_func != 'shift' and shift_func != 'roll':
        raise KeyError('Shift function has to be either shift or roll')

    image_f = np.fft.rfft2((image - np.mean(image)) * real_filter)
    reference_f = np.fft.rfft2((reference - np.mean(reference)) * real_filter)

    xcor = abs(np.fft.irfft2(np.conj(image_f) * reference_f * k_filter))
    if verbose:
        import matplotlib.pyplot as plt
        plt.imshow(np.fft.fftshift(xcor))
        plt.title('imageCrossCorRealShift xcor')
    shifts = np.unravel_index(np.fft.fftshift(xcor).argmax(), xcor.shape)
    shifts = (shifts[0] - xcor.shape[0] / 2, shifts[1] - xcor.shape[1] / 2)
    shifts = [int(i) for i in shifts]  # convert to integers

    if shift_func == 'shift':
        # shift image using ndimage.shift
        output = ndimage.shift(image, shifts, order=0)
    elif shift_func == 'roll':
        # shift image using roll to be reversible
        output = np.roll(image, shifts[0], axis=0)
        output = np.roll(output, shifts[1], axis=1)

    return output, shifts


def image_phase_correlate(image, reference, real_filter=1, k_filter=1, shift_func='shift', verbose=False):
    """ Align image to reference by phase-correlation. Outputs shifted images and shift.
    Uses np.fft.rfft2 for ~2x speed improvement.
    Uses scipy.ndimage.shift() to shift the image and remove border pixels.

    NOT WORKING OR TESTED YET

    Note
    ----
        image, reference and real_filter must all have the same shape (N, M).
        k_filter must have a shape that matches the np.fft.rfft2() of
        the other inputs: (N, M/2+1)

    Parameters
    ----------
        image : ndarray
            A image as a 2D ndarray.

        reference : ndarray
            The reference image to align to.

        real_filter : ndarray, optional, default = 1
            A real space filter applied to image and reference before
            calculating the shift.

        k_filter : ndarray, optional, default = 1
            A Fourier space filter applied to the fourier transform of
            image and reference before calculating the cross-correlation.

        shift_func : str, default is 'shift'
            The function to use to shift the images. 'roll' uses np.roll and 'shift' uses ndimage.shift.


        verbose : bool
            Plots the cross-correlation using matplotlib for debugging purposes.

    Returns
    ------
        : tuple, (ndarray, tuple)
            A tuple containing the shifted image and the shifts applied.

    """
    output = None

    image_f = np.fft.rfft2((image - np.mean(image)) * real_filter)
    reference_f = np.fft.rfft2((reference - np.mean(reference)) * real_filter)

    xcor = abs(np.fft.irfft2(np.conj(image_f) * reference_f * k_filter))
    pcor = xcor / (np.abs(xcor) + 0.001)
    if verbose:
        import matplotlib.pyplot as plt
        plt.imshow(np.fft.fftshift(pcor))
        plt.title('imageCrossPhaseRealShift pcor')
    shifts = np.unravel_index(np.fft.fftshift(pcor).argmax(), pcor.shape)
    shifts = (shifts[0] - pcor.shape[0] / 2, shifts[1] - pcor.shape[1] / 2)
    shifts = [int(i) for i in shifts]  # convert to integers

    if shift_func == 'shift':
        # shift image using ndimage.shift
        output = ndimage.interpolation.shift(image, shifts, order=0)
    elif shift_func == 'roll':
        # shift image using roll to be reversible
        output = np.roll(image, shifts[0], axis=0)
        output = np.roll(output, shifts[1], axis=1)

    return output, shifts


def stack_align(stack, align_type='static', real_filter=1, k_filter=1, shift_func='shift'):
    """ Align a series of images by cross-correlation. All images are aligned to the first image Uses image_correlate
    which is based on simple cross correlation.

    Notes
    -----
        You should probably use ncempy.algo.stack_align since it uses mutlicorr and is more
        functional.

    Parameters
    ----------
        stack : ndarray, 3D
            The stack of images to align. Shape [num, Y, X]

        real_filter : ndarray, optional, default = 1
            A real space filter to apply before cross-correlation
            of each image. Shape must be [Y, X]

        k_filter : ndarray, optional, default = 1
            A Fourier space filter to apply before cross-correlation.
            Shape must be [Y, X/2 + 1]

        shift_func : str, default is 'shift'
            The function to use to shift the images. 'roll' uses np.roll and 'shift' uses ndimage.shift.

        align_type: str
            static or dynamic alignment. Static aligns all images to the first image. Dynamic aligns
            each image to the previous image starting with the first image

    Returns
    -------
        : tuple, aligned stack, shifts
            A tuple containing the aligned images as a 3D ndarray of shape
            [num, Y, X] and shifts as a 2D ndarray of shape [num, 2]
    """

    if align_type != 'static' and align_type != 'dynamic':
        raise KeyError('Incorrect align type. Must be static or dynamic')

    # Pre-allocate the arrays
    aligned = np.zeros_like(stack)  # shifted data array
    shifts = np.zeros((stack.shape[0], 2))  # the applied shifts

    aligned[0, :, :] = stack[0, :, :]

    jj = 0
    ref_sh = np.zeros((2,))
    for ii in range(1, stack.shape[0]):
        output, sh = image_correlate(stack[ii, :, :], stack[jj, :, :], real_filter, k_filter, shift_func=shift_func)
        sh += ref_sh
        aligned[ii, :, :] = output
        shifts[ii, :] = sh
        if align_type == 'dynamic':
            ref_sh = sh
            jj = ii

    return aligned, shifts

def moments(im, order=3):
    """ Calculate the image raw moments.

    Parameters
    ----------
        im : ndarray
            A 2D array of the image to calculate the moments for

        order : int, default=3
            The order to calculate the moments to.

    Returns
    -------
        : ndarray
            An ndarray of (order+1, order+1)
    """
    order = int(order) + 1
    XX, YY = np.mgrid[0:im.shape[0], 0:im.shape[1]]
    mm = np.zeros((order, order))
    for ii in range(order):
        XXii = XX ** ii
        for jj in range(order):
            mm[ii, jj] = np.sum(XXii * YY ** jj * im.astype(np.float64))
    return mm


def moments_central(im, cent=None, order=3):
    """ Calculate the image central moments.

    Parameters
    ----------
        im : ndarray
            A 2D array of the image to calculate the central moments for

        cent : tuple, 2 values, optional
            The centroid calculated from the image raw moments. If it
            is not supplied then it is calculated by moments()

        order : int, default = 3
            The order to calculate the orders to.

    Returns
    -------
        : ndarray
            An ndarray of size (order + 1, order +1)

    """
    if not cent:
        m = moments(im, order)
        cent = centroid(m)

    order = int(order) + 1

    XX, YY = np.mgrid[-cent[0]:im.shape[0] - cent[0],
                      - cent[1]:im.shape[1] - cent[1]]
    XX = XX.astype('float64')
    YY = YY.astype('float64')

    mc = np.zeros((order, order))
    for ii in range(order):
        XXii = XX ** ii
        for jj in range(order):
            mc[ii, jj] = np.sum(XXii * YY ** jj * im.astype(np.float64))
    return mc


def centroid(m):
    """ Find the centroid based on the raw moments.

    Parameters
    ----------
        m : ndarray
            The output from moments()

    Returns
    -------
        : tuple, 2 elements (X,Y)
            The position of the centroid as a 2-tuple
    """
    return m[1, 0] / m[0, 0], m[0, 1] / m[0, 0]


def moment_angle(mc):
    """ Calculate the orientation of the object from its central moments

    Parameters
    ----------
        mc : ndarray
            The array of the central moments. Output from moments_central

    Returns
    -------
        : float
            The angle in radians.

    """
    mu_20 = mc[2, 0] / mc[0, 0]  # - centroid[0]**2
    mu_02 = mc[0, 2] / mc[0, 0]  # - centroid[1]**2
    mu_11 = mc[1, 1] / mc[0, 0]  # - centroid[0]*centroid[1]

    th = 0.5 * np.arctan2(2 * mu_11, mu_20 - mu_02)
    return th

def shearImage(im, dim, shear_factor):
    """ Exact shear of an image using the frequency space of the FFT
    of the image.

    Currently only works for square images.

    Parameters
    ----------
        im: ndarray, 2D
            The image to shear.
        dim: int
            The axis to shear along
        shear_factor: float
            The amount of shear.

    Returns
    -------
        : ndarray
            The sheared image.
    """

    if im.shape[0] != im.shape[1]:
        print('Input image must be square.')
        return

    fftF = np.fft.fftshift(np.fft.fftfreq(im.shape[dim], d=1 / im.shape[dim]))
    yy, xx = np.meshgrid(fftF, fftF, indexing='ij')
    f = np.fft.fftshift(np.fft.fftn(im, axes=(dim,)), axes=(dim,))
    ff = f * np.exp(-2.0j * np.pi / im.shape[dim] * xx * yy * shear_factor)

    sh = np.fft.ifftn(np.fft.ifftshift(ff, axes=(dim,)), axes=(dim,))
    sh = np.real(sh)
    return sh


def shiftImage(im, shift):
    """ Exact shear of an image using the frequency space of the FFT
    of the image.

    Parameters
    ----------
        im: ndarray, 2D
            The image to shear.
        shift: tuple of floats
            The number of pixels to shift in each direction
    Returns
    -------
        : ndarray
            The shifted image.
    """

    fftF0 = np.fft.fftshift(np.fft.fftfreq(im.shape[0], d=1 / im.shape[0]))
    fftF1 = np.fft.fftshift(np.fft.fftfreq(im.shape[1], d=1 / im.shape[1]))
    xx, yy = np.meshgrid(fftF0, fftF1, indexing='ij')
    f = np.fft.fftshift(np.fft.fftn(im))
    ff = f * np.exp(-2.0j * np.pi / im.shape[0] * xx * shift[0] +
                    -2.0j * np.pi / im.shape[1] * yy * shift[1])

    sh_im = np.fft.ifftn(np.fft.ifftshift(ff))
    sh_im = np.real(sh_im)
    return sh_im


def rotateImage(im, theta, pad=False):
    """ Use three shears in Fourier space to exactly (and reversibly) rotate
    an image.

    Currently only works for square images.

    Parameters
    ----------
        im: ndarray, 2D
            The image to rotate

        theta: float
            The angle to rotate by in radians

        pad: bool
            Add padding to the image before rotating

    """

    if im.shape[0] != im.shape[1]:
        print('Input image must be square.')
        return

    alpha = gamma = - np.tan(theta/2.)
    beta = np.sin(theta)

    if pad:
        p = im.shape[0] * np.sin(theta)
        im = np.pad(im, int(p//2+1))

    im1 = shearImage(im,  0, alpha)
    im2 = shearImage(im1, 1, beta)
    im3 = shearImage(im2, 0, gamma)
    im3 = np.real(im3)
    return im3



"""
Module to find local maxima in 2D and 3D data such as images and density maps.

Many functions are based on work by Colin Ophus, cophus@lbl.gov and his excellent
RealSpaceLattice01.m code.

author: Peter Ercius, percius@lbl.gov
"""

try:
    from . import gaussND
except:
    try:
        print('peakFind: Temporary import!')
        import gaussND
    except ModuleNotFoundError:
        raise ModuleNotFoundError
    

import numpy as np
import scipy.optimize as opt
from scipy import ndimage

default_indexing: str = 'ij'


def doubleRoll(image, vec):
    """ Roll an 2D ndarray (e.g. an image) along each dimension. 
    
        Parameters
        ----------
            image : np.ndarray
                The 2D array to roll.
            
            vec : iterable
                The number of time to apply the roll to the rows and columns where
                the tuple has the structure (row, col). Values in tuple must be int.
                
        Returns
        -------
            : np.ndarray
                The rolled image.
    """
    assert len(vec) == 2
    return np.roll(np.roll(image, vec[0], axis=0), vec[1], axis=1)


def tripleRoll(vol, vec):
    """ Roll a 3D ndarray (e.g. an volume) along each dimension. 
    
        Parameters
        ----------
            vol : np.ndarray
                The 3D array to roll
            
            vec : iterable
                The number of time to apply the roll the tuple has the structure (axis0, axis1, axis2).
                Values in tuple must be int.
                
        Returns
        -------
            : ndarray
                The rolled volume.
    """
    assert len(vec) == 3
    return np.roll(np.roll(np.roll(vol, vec[0], axis=0), vec[1], axis=1), vec[2], axis=2)


def peakFind2D(image, threshold):
    """ Find peaks in a 2D image based on the roll technique. A threshold is first applied to remove
    low intensity noisy peaks.
    
    Parameters
    ----------
        image : ndarray
            An 2D ndarray of image intensities with peaks to find.
        
        threshold : float
            Threshold [0,1) the image to remove noisy peaks based on max intensity in the image
        
    Returns
    -------
        positions : array
            A Nx2 array with the index locations of the n number of peaks
    """
    if (threshold < 0) or (threshold >= 1):
        print('Error: Threshold must be 0 <= threshold < 1')
        return 0

    # Could generate the set of coordinates using
    # list(set(permutations((0,1,1,-1,-1),2)))
    pLarge = (image > doubleRoll(image, [-1, -1])) & \
             (image > doubleRoll(image, [0, -1])) & \
             (image > doubleRoll(image, [1, -1])) & \
             (image > doubleRoll(image, [-1, 0])) & \
             (image > doubleRoll(image, [1, 0])) & \
             (image > doubleRoll(image, [-1, 1])) & \
             (image > doubleRoll(image, [0, 1])) & \
             (image > doubleRoll(image, [1, 1])) & \
             (image > threshold * np.max(image))

    positions = np.array(np.where(pLarge * image)).T.copy()  # return array as [num,posXY]
    return positions.astype(int)


def peakFind3D(vol, threshold):
    """
    Find peaks in a 3D volume based on the roll technique. A threshold is first applied to remove
    low intensity noisy peaks.
    
    Parameters
    ------
        vol : ndarray
            An 3D ndarray of image intensities with peaks to find.

        threshold : float
            Threshold [0,1) the image to remove noisy peaks based on max intensity in the image
        
    Returns
    -------
        positions : ndarray
            A Nx3 array with the index locations of the n number of peaks
    """
    if (threshold < 0) or (threshold >= 1):
        print('Error: Threshold must be 0 <= threshold < 1')
        return 0

    # Could generate the set of coordinates programmatically
    # using list(set(permutations((0,0,1,1,1,-1,-1,-1),3))
    pLarge = ((vol > tripleRoll(vol, [-1, -1, -1]))
              & (vol > tripleRoll(vol, [0, -1, -1]))
              & (vol > tripleRoll(vol, [1, -1, -1]))
              & (vol > tripleRoll(vol, [-1, 0, -1]))
              & (vol > tripleRoll(vol, [1, 0, -1]))
              & (vol > tripleRoll(vol, [-1, 1, -1]))
              & (vol > tripleRoll(vol, [0, 1, -1]))
              & (vol > tripleRoll(vol, [1, 1, -1]))
              & (vol > tripleRoll(vol, [0, 0, -1]))
              & (vol > tripleRoll(vol, [-1, -1, 0]))
              & (vol > tripleRoll(vol, [0, -1, 0]))
              & (vol > tripleRoll(vol, [1, -1, 0]))
              & (vol > tripleRoll(vol, [-1, 0, 0]))
              & (vol > tripleRoll(vol, [1, 0, 0]))
              & (vol > tripleRoll(vol, [-1, 1, 0]))
              & (vol > tripleRoll(vol, [0, 1, 0]))
              & (vol > tripleRoll(vol, [1, 1, 0]))
              & (vol > tripleRoll(vol, [-1, -1, 1]))
              & (vol > tripleRoll(vol, [0, -1, 1]))
              & (vol > tripleRoll(vol, [1, -1, 1]))
              & (vol > tripleRoll(vol, [-1, 0, 1]))
              & (vol > tripleRoll(vol, [1, 0, 1]))
              & (vol > tripleRoll(vol, [-1, 1, 1]))
              & (vol > tripleRoll(vol, [0, 1, 1]))
              & (vol > tripleRoll(vol, [1, 1, 1]))
              & (vol > tripleRoll(vol, [0, 0, 1]))
              & (vol > threshold * np.max(vol)))

    positions = np.array(np.where(pLarge * vol)).T.copy()  # return array as [num,posXYZ]
    return positions.astype(int)


def enforceMinDist(positions, intensities, minDistance):
    """ Remove peaks that violate a minimum distance requirement. If two peaks are
    closer together than the minDistance then the one with the highest intensity is chosen.
    
    Works for 2D and 3D data sets.
    
    Parameters
    ----------
        positions : np.ndarray
            An array of shape Nx3 or Nx2 containing the positions of the peaks to test 
        intensities : np.ndarray
            An array of shape Nx1 with the peak intensities.
        minDistance : int
            The minimum distance in pixels
        
    Returns
    -------
        validPeaks : np.ndarray
            An ndarray of Mx3 or Mx2 (M <= N) with invalid peaks removed.
    """

    validPeaks = np.zeros(positions.shape, dtype=np.float32)  # Will hold intensity,X,Y,Z

    alreadyTested = np.zeros(
        positions.shape[0])  # all peaks are tested unless they were already compared to another peak

    for ii, p0 in enumerate(positions):  # ii is the iteration number and p0 is the element
        if not alreadyTested[ii]:
            pointsDist = np.sqrt(np.sum((positions - p0) ** 2, axis=1))  # find the distance to every other point
            closePeaks = pointsDist[(pointsDist > 0) & (pointsDist < minDistance)]

            if closePeaks.any():  # test for any values < minDistance
                closePeaksIndices = np.where(pointsDist < minDistance)[0]  # find all close peaks including self
                peakIntensities = intensities[closePeaksIndices]
                highestPeakIndex = closePeaksIndices[np.argmax(peakIntensities)]  # find the index of the highest peak
                alreadyTested[closePeaksIndices] = 1  # Don't test peaks in the future if already tested
                validPeaks[closePeaksIndices, :] = np.nan  # set all values to nan; cant be integer data set
                validPeaks[ii, :] = positions[highestPeakIndex, :]  # set highest intensity to correct values
            else:
                validPeaks[ii, :] = positions[ii, :]  # save the current peak
                # alreadyTested[ii] = 1 #Dont test peaks in the future if already tested

    validPeaks = validPeaks[~np.isnan(validPeaks[:, 0]), :]
    # print("Removed %d peaks" % (positions.shape[0] - validPeaks.shape[0]))
    return validPeaks.astype(int)


def peaksToVolume(peakList, volShape, gaussSigma, gaussSize, indexing='ij'):
    """ Place 3D Gaussian at a set of peak positions.
    
    Parameters
    ----------
        peakList : np.ndarray
            Array of peak positions of size (numPeaks, 3).
        
        volShape : tuple
            3 element tuple of the shape of the volume containing the peaks
        
        gaussSigma : tuple
            3 element tuple with that sigma parameters passed to the gaussND.gauss3D() function (sigX,sigY,sigZ)
        
        gaussSize : tuple
            3 element tuple describing the 3D size of the simulated gaussian box. Must be odd.

        indexing : str
            String to pass to np.meshgrid to indicate which indexing is used. Either 'ij' or 'xy' are possible. Default
            is 'ij' for this module.
    Returns
    -------
        : np.ndarray
            A volume containing gaussian peaks at each position indicated in peakList parameter.
    
    """

    sim_volume = np.zeros(volShape, dtype=np.float32)

    if np.any(np.array(gaussSize) % 2 == 0):
        raise ValueError('gaussSize must be odd.')

    sz = [int((g - 1) // 2) for g in gaussSize]
    temp0 = np.arange(-sz[0], sz[0]+1)
    temp1 = np.arange(-sz[1], sz[1]+1)
    temp2 = np.arange(-sz[2], sz[2]+1)
    Y3D, X3D, Z3D = np.meshgrid(temp0, temp1, temp2, indexing=indexing)

    for pos in peakList:
        pos_loc = (pos // 1).astype(int)

        pos_remain = pos % 1
        gg3 = gaussND.gauss3D(Y3D, X3D, Z3D,
                              pos_remain[0], pos_remain[1], pos_remain[2],
                              gaussSigma[0], gaussSigma[1], gaussSigma[2])

        sim_volume[pos_loc[0] - sz[0]:pos_loc[0] + sz[0] + 1,
                   pos_loc[1] - sz[1]:pos_loc[1] + sz[1] + 1,
                   pos_loc[2] - sz[2]:pos_loc[2] + sz[2] + 1] += gg3
    return sim_volume


def peaksToImage(peakList, imShape, gaussSigma, gaussSize, indexing='ij'):
    """ Place 2D Gaussian at a set of peak positions.
    
    Parameters
    ----------
        peakList : np.ndarray
            Array of peak positions of size (numPeaks, 2).
        
        imShape : tuple
            2 element tuple of the shape of the image containing the peaks
        
        gaussSigma : tuple
            2 element tuple with that sigma parameters passed to the gaussND.gauss2D() function (sigX,sigY)
        
        gaussSize : tuple
            2 element tuple describing the 2D size of the simulated gaussian box. Must be odd.

        indexing : str, default = 'ij'
            'ij' or 'xy' indexing to pass to np.meshgrid.
    Returns
    -------
        : np.ndarray
            A image containing Gaussian peaks at each position indicated in peakList parameter.
    
    """

    sim_image = np.zeros(imShape, dtype=np.float32)

    if np.any(np.array(gaussSize) % 2 == 0):
        raise ValueError('gaussSize must be odd.')

    sz = [int((g - 1) // 2) for g in gaussSize]
    temp0 = np.arange(-sz[0], sz[0] + 1)
    temp1 = np.arange(-sz[1], sz[1] + 1)
    Y3D, X3D = np.meshgrid(temp1, temp0, indexing=indexing)
    for pos in peakList:
        pos_loc = (pos // 1).astype(int)
        pos_remain = pos % 1
        gg2 = gaussND.gauss2D(Y3D, X3D,
                              pos_remain[0], pos_remain[1],
                              gaussSigma[0], gaussSigma[1])
        sim_image[pos_loc[0] - sz[0]:pos_loc[0] + sz[0] + 1,
                  pos_loc[1] - sz[1]:pos_loc[1] + sz[1] + 1] += gg2
    return sim_image


def lattice2D_norm(u, v, a, b, origin, num_points):
    """
    Returns a set of points in a lattice according to the u, v unit vectors (vectors are normalized internally)
    and lengths a,b centered at origin. The lattice has num_points along each u,v
    vector.
    
    Parameters
    ----------
        u, v : 2-tuple
            2 element tuples defining the lattice directions as vectors.

        a, b : float
            values to multiply each vector by (if u,v are not normalized then set these to 1)

        origin : 2-tuple
            The origin in the format (x0, y0)

        num_points : 2-tuple
            2 element tuple of the number of repeats of the lattice along each direction

    Returns
    ------
        : ndarray
            The set of points in the lattice. Size (num_points[0] * num_points[1], 2)
    """

    assert np.linalg.norm(u) == 1
    assert np.linalg.norm(v) == 1

    totalNumPoints = np.prod(num_points)

    Y2D, X2D = np.meshgrid(range(num_points[1]), range(num_points[0]), indexing=default_indexing)

    X = X2D.reshape(totalNumPoints)
    Y = Y2D.reshape(totalNumPoints)

    xy = np.zeros((np.prod(num_points), 2))
    xy[:, 0] = origin[0] + a * X * u[0] + b * Y * v[0]
    xy[:, 1] = origin[1] + a * X * u[1] + b * Y * v[1]

    return xy


def lattice2D(u, v, a, b, origin, num_points):
    """
    A modified version of peakFind.lattice2D_norm which can use non normalized u,v,vectors
    
    Parameters
    ----------
        u, v : tuple
            2 element vectors defining the lattice directions

        a, b : float
            values to multiply each vector by (if u,v,w are not normalized then set these to 1)

        origin : tuple
            The origin in the format (x0,y0)

        num_points : tuple
            2 element tuple of the number of repeats of the lattice along each direction

    Returns
    ------
        : ndarray
            The set of points in the lattice. Size (num_points[0] * num_points[1], 2)
    """
    totalNumPoints = np.prod(num_points)

    Y2D, X2D = np.meshgrid(range(num_points[1]), range(num_points[0]), indexing=default_indexing)

    X = X2D.reshape(totalNumPoints)
    Y = Y2D.reshape(totalNumPoints)

    xy = np.zeros((np.prod(num_points), 2))
    xy[:, 0] = origin[0] + a * X * u[0] + b * Y * v[0]
    xy[:, 1] = origin[1] + a * X * u[1] + b * Y * v[1]

    return xy


def lattice3D(u, v, w, a, b, c, origin, num_points):
    """
    Returns a set of points in a lattice according to the u, v, w vectors
    and lengths a,b centered at origin. The lattice has num_points along each u,v,w
    vector.

    Parameters
    ----------
        u, v, w : tuple or np.ndarray
            3 element vectors defining the lattice directions
        a, b, c : float
            Values to multiply each vector by (if u,v,w are not normalized then set these to 1)
        origin : tuple
            The origin in the format (x0,y0,z0)
        num_points : tuple
            3 element tuple of the number of repeats of the lattice along each direction

    Returns
    -------
        xyz : ndarray
            A (N,3) shaped set of coordinates
    """

    totalNumPoints = np.prod(num_points)

    Y3D, X3D, Z3D = np.meshgrid(range(num_points[1]), range(num_points[0]), range(num_points[2]),
                                indexing=default_indexing)

    X = X3D.reshape(totalNumPoints)
    Y = Y3D.reshape(totalNumPoints)
    Z = Z3D.reshape(totalNumPoints)

    xyz = np.zeros((np.prod(num_points), 3))
    xyz[:, 0] = origin[0] + a * X * u[0] + b * Y * v[0] + c * Z * w[0]
    xyz[:, 1] = origin[1] + a * X * u[1] + b * Y * v[1] + c * Z * w[1]
    xyz[:, 2] = origin[2] + a * X * u[2] + b * Y * v[2] + c * Z * w[2]

    return xyz


def applyLatticeLimit(lattice, bounds):
    """Remove lattice points outside the data bounds. For 2D and 3D data.
    
    Parameters
    ---------
        lattice : ndarray; (N, 2) or (N, 3)
            From lattice2D

        bounds : tuple, 
            Minimum and maximum for axes 0 and 1
            (min0, max0, min1, max1) or axes 0, 1 and 2
            (min0, max0, min1, max1, min2, max2)

    Returns
    -------
        : ndarray; (M, 2) or (M, 3)
            Same as lattice input except only containing
            points within the bounds specified. M <= N
    """
    if len(bounds) == 4:
        goodUVs = (((lattice[:, 0] > bounds[0]) & (lattice[:, 0] < bounds[1])) &
                   ((lattice[:, 1] > bounds[2]) & (lattice[:, 1] < bounds[3])))
    elif len(bounds) == 6:
        goodUVs = (((lattice[:, 0] > bounds[0]) & (lattice[:, 0] < bounds[1])) &
                   ((lattice[:, 1] > bounds[2]) & (lattice[:, 1] < bounds[3])) &
                   ((lattice[:, 2] > bounds[4]) & (lattice[:, 2] < bounds[5])))
    else:
        print('Bounds needs be be either 4 or 6 value tuple.')
        return None
    return lattice[goodUVs, :]


# noinspection PyArgumentList
def peakPlot3D(X, Y, Z, mkr, myAxes3D):
    """
    Plot a set of peaks in a 3D plot using matplotlib. See the example below for how to set up a figure as input
    to this function using matplotlib Axes3D.

    todo: move this to viz

    Parameters
    -----------
        X : np.ndarray
            The x positions as a 1D array.

        Y : np.ndarray
            The y positions as a 1D array.

        Z : np.ndarray
            The Z positions as a 1D array.

        mkr : string
            String indicating the marks type and color. Passed directly to pyplot.plt
            command (ex. 'go' for green circles)

        myAxes3D : mpl_toolkits.mplot3d.Axes3d
            An Axes3D object. (see example below)
        
    Returns
    -------
        None
    
    Example
    -------
    This function requires the input of a Axes3D to plot a set of peaks. See below how to import
    and set up a figure for use with this function.
    >> import matplotlib.pyplot as plt
    >> import mpl_toolkits.mplot3d
    >> from ncempy.algo import peak_find
    >> fg1 = plt.figure()
    >> ax1 = mpl_toolkits.mplot3d.Axes3D(fg1)
    >> peak_find.peakPlot3D(peakList[:,2], peakList[:,1], peakList[:,0], 'go', ax1)
    """

    import matplotlib.pyplot as plt

    plt.plot(X, Y, Z, mkr)

    max_range = np.array([X.max() - X.min(), Y.max() - Y.min(), Z.max() - Z.min()]).max() / 2.0

    mid_x = (X.max() + X.min()) * 0.5
    mid_y = (Y.max() + Y.min()) * 0.5
    mid_z = (Z.max() + Z.min()) * 0.5

    myAxes3D.set_xlim(mid_x - max_range, mid_x + max_range)
    myAxes3D.set_ylim(mid_y - max_range, mid_y + max_range)
    myAxes3D.set_zlim(mid_z - max_range, mid_z + max_range)

    myAxes3D.set_xlabel('X')
    myAxes3D.set_ylabel('Y')
    myAxes3D.set_zlabel('Z')


def peak3View(fg, vol, peak_positions):
    """
    Plot 3 orthogonal slices and the corresponding peaks in each slice.
    
    Not fully test. Unsure whether the slices and peaks are exactly the same.

    todo: move this to viz

    Parameters
    -----------
        fg : matplotlib figure
            The figure to use. This allows the user to reuse figures

        vol : ndarray (3D)
            The volume to slice along each orthogonal direction. Set up for [X,Z,Y] orientation

        peak_positions : ndarray, list, tuple (1D)
            This selects the point shown in all three slices
        
    Returns
    -------
        None

    """
    ((ax1, ax2), (ax3, ax4)) = fg.subplots(2, 2)

    slices = [int(round(ii)) for ii in peak_positions]
    ax1.imshow(vol[slices[0], :, :])
    ax1.plot(peak_positions[2], peak_positions[1], 'xk')
    ax1.set(title='X slice {}'.format(slices[0]))

    ax3.imshow(vol[:, slices[1], :])
    ax3.plot(peak_positions[2], peak_positions[0], 'xk')
    ax3.set(title='Y slice {}'.format(slices[1]))

    ax4.imshow(vol[:, :, slices[2]])
    ax4.plot(peak_positions[1], peak_positions[0], 'xk')
    ax4.set(title='Z slice {}'.format(slices[2]))

    ax2.axis('off')
    fg.tight_layout()


def remove_xrays(imageOriginal, threshold, size_median_filter=(3, 3)):
    """
    Find x-ray pixels in TEM images and replace the xray intensity 
    with a local median value.
    
    Parameters
    ----------
        imageOriginal : ndarray
            The image as a 2D ndarray.
        threshold : float
            Value above which to consider a pixel intensity an x-ray.
        size_median_filter : list or tuple
            The footprint of the median filter over which to search for a median value.
            
    Returns
    -------
        : ndarray
            The image with x-ray values replaced with median local value.
    
    """
    # Find x-ray pixels
    image_med_filter = ndimage.filters.median_filter(imageOriginal, size_median_filter)
    sub_xrays = np.where(imageOriginal > (image_med_filter + threshold))

    # Replace x-ray pixels with median filtered values.
    image_filtered = imageOriginal.copy()
    image_filtered[sub_xrays[0], sub_xrays[1]] = image_med_filter[sub_xrays[0], sub_xrays[1]]
    return image_filtered, sub_xrays


def writeXYZ(filename, XYZ, element, comment):
    """
    Write out a set of XYZ coordinates that can be read by various crystal viewing software such as Vesta.

    todo: Move this to io

    Parameters
    ----------
        filename : str
            The name of the file to write to.

        XYZ : np.ndarray
            Atom coordinates in a 2D ndarray with axes [atom number, position]. Should be shape [N, 3]

        element : tuple
            Element symbol of each coordinate as a string (e.g. 'C'). Should be length N.

        comment : str
            A comment to add to the file.

    Example
    -------
        # Write out the peak positions as Carbon atoms:
        >> writeXYZ('name.xyz', positions, ['C',] * len(positions), 'writeXYZ example carbon atoms')
    """
    numAtoms = XYZ.shape[0]
    with open(filename, 'w') as f1:
        f1.write('{}\n'.format(numAtoms))
        f1.write('{}\n'.format(comment))
        for ii in range(numAtoms):
            f1.write('{} {} {} {} \n'.format(element[ii], XYZ[ii, 0], XYZ[ii, 1], XYZ[ii, 2]))


def refineLattice2D(or0, u0, v0, pos, fraction=(1, 1), max_iter=30,
                    refine_locally=True, verbose=False, num_unit_cells=(1, 1)):
    """ Refine lattice based on measurements and initial
    guess at lattice vectors. This code is designed to work only with
    square lattices. Hexagonal and more complex lattice might not work.

    Parameters
    ----------
        or0 : tuple
            Initial guess of origin. 2-tuple
        u0 : tuple
            Initial u vector. Use num_unit_cells to average over several lengths of the lattice vector in this
            direction. 2-tuple
        v0 : tuple
            Initial v vector. Use num_unit_cells to average over several lengths of the lattice vector in this
            direction. 2-tuple
        pos : np.ndarray
            The set of positions to fit the lattice to. (number, position) (M, 2)
        refine_locally : bool
            Refine locally near the origin before using all positions
            Locally is considered 2 time the larger vector of u0 or v0.
        fraction : tuple
            Site fraction to take into account peak positions inside the unit cell. FCC imaged along [100] for\
            example would need site_fraction = (2, 2).
        max_iter : int
            The maximum number of iterations to run to refine. This usually converges in a few iterations. Default is
            30.
        num_unit_cells : tuple
            The number of unit cells the initial guess of u0 and v0 extend over. 2-tuple (num_u0, num_v0) with default
        verbose : bool
            Print out helpful information.
            
    Returns
    -------
        : tuple
            Tuple of of 4 np.ndarray (origin, u, v, ab) where ab is the
            site positions in fractions of u and v.
    
    """

    origin = or0
    u = u0
    v = v0

    xyRes = (0, 0)
    ab = (0, 0)

    # Divide by number plotted
    u = [ii / num_unit_cells[0] for ii in u]
    v = [ii / num_unit_cells[1] for ii in v]

    for ii in range(max_iter):
        if refine_locally is True:
            for jj in range(1, 30):
                # Refine locally first stepping out
                rMax = jj * np.mean((np.linalg.norm(u), np.linalg.norm(v)))  # radius to use first

                # compute a, b, c values by finding points p0 in basis {u,v,w} from {i,j,k}
                uv = np.array([u, v]).T
                pR = np.sqrt((pos[:, 0] - origin[0]) ** 2 + (pos[:, 1] - origin[1]) ** 2)
                p1 = pos[pR < rMax, :]
                if p1.shape[0] < 10:
                    # Test that there are enough atoms to do a fit. Skip if less than 10
                    continue
                (ab, abRes, abX, abY) = np.linalg.lstsq(uv, (p1 - origin).T, rcond=None)

                # Refine lattice
                A = np.ones([ab.shape[1], 3])  # origin values should be 1
                A[:, 1] = np.round(fraction[0] * ab[0, :]) / fraction[
                    0]  # fill the rest of the values with the abc values
                A[:, 2] = np.round(fraction[1] * ab[1, :]) / fraction[1]

                (xy_beta, xyRes, rXY, sXY) = np.linalg.lstsq(A, p1, rcond=None)

                origin = xy_beta[0, :]
                u = xy_beta[1, :]
                v = xy_beta[2, :]

        uv = np.array([u, v]).T
        (ab, abRes, abX, abY) = np.linalg.lstsq(uv, (pos - origin).T, rcond=None)

        # Refine lattice
        A = np.ones([ab.shape[1], 3])
        A[:, 1] = np.round(fraction[0] * ab[0, :]) / fraction[0]  # fill the rest of the values with the abc values
        A[:, 2] = np.round(fraction[1] * ab[1, :]) / fraction[1]
        (xy_beta, xyRes, rXY, sXY) = np.linalg.lstsq(A, pos, rcond=None)

        origin = xy_beta[0, :]
        u = xy_beta[1, :]
        v = xy_beta[2, :]
        if verbose:
            print('[u,v] = [[{}],[{}]]'.format(u, v))

    if verbose:
        print('Residuals = [{}, {}]'.format(xyRes[0], xyRes[0]))
        # Print out the results
        print('origin: {}'.format(origin))
        print('u: {}\nv: {}'.format(u, v))
        print('|u|: {0[0]:.4}, |v|: {0[1]:.4} (pixels)'.format(np.linalg.norm([u, v], axis=1)))
        print('angle UV: {:.4}'.format(np.arccos(np.dot(u / np.linalg.norm(u),
                                                        v / np.linalg.norm(v))) * 180. / np.pi))

    return origin, u, v, ab.T


def refineLattice3D(or0, u0, v0, w0, pos, fraction=(1, 1, 1), max_iter=30, refine_locally=True):
    """ Refine lattice based on measurements and initial guess at lattice vectors
    
    Warning
    -------
        Tested code but not production yet. This is a work un progress.
    
    Parameters
    ----------
        or0 : 3-element tuple 
            Origin.
        u0 : tuple
            Initial u vector.    
        v0 : tuple 
            Initial v vector
        w0 : tuple 
            Initial w vector
        pos : array (number, position) (M,3)
            The set of positions to fit the lattice to.
        refine_locally : bool
            Refine locally near the origin before using all positions
            Locally is considered 2 time the larger vector of u0 or v0.
        fraction : tuple
            Site fraction to take into account peak positions inside the unit cell. FCC imaged along [100] for
            example would need site_fraction = (2, 2).
        max_iter : int
            Maximum number of iterations.
    Returns
    -------
        : tuple of 4 np.ndarray (origin, u, v, ab)
            Tuple of refined origin, u, v optimized values as np.ndarray and the
            site positions in fractions of u and v.
    
    """
    xRes0 = 1e6
    yRes0 = 1e6
    zRes0 = 1e6
    for ii in range(max_iter):

        # compute a,b values
        uv = np.array([u0, v0, w0]).T
        (ab, abRes, abX, abY) = np.linalg.lstsq(uv, (pos - or0).T, rcond=None)

        # Refine lattice
        A = np.ones([ab.shape[1], 4])
        A[:, 1] = np.round(ab[0, :])
        A[:, 2] = np.round(ab[1, :])
        A[:, 3] = np.round(ab[2, :])
        (xy_beta, xyRes, rXY, sXY) = np.linalg.lstsq(A, pos, rcond=None)

        if xyRes[0] < xRes0 and xyRes[1] < yRes0 and xyRes[2] < zRes0:
            _ = xyRes[0]
            _ = xyRes[1]
            _ = xyRes[2]
        else:
            print('----Converged----')
            break

        print('Residuals = [{0[0]}, {0[1]}, {0[2]}]'.format(xyRes))

        origin = xy_beta[0, :]
        u = xy_beta[1, :]
        v = xy_beta[2, :]
        w = xy_beta[3, :]

        return origin, u, v, w


def generateLatticeFromRefinement(origin, u, v, ab, fraction=(1, 1)):
    """ Generate lattice positions from refined output
    of refineLattice2D. This can be used to generate
    all positions when fraction is used in refineLattice2D
    
    Parameters
    ----------
        origin : tuple
            Origin. 2-tuple
        u : tuple
            Initial u vector. 2-tuple
        v : tuple
            Initial v vector. 2-tuple
        fraction : tuple
            The fraction used in refineLattice2D. 2-tuple
        ab : np.ndarray
            The set of positions in terms of u and v lattice vectors. (number, position) (M, 2)
        fraction : tuple
            The fractional coordinates for the unit cell. 2-tuple
    
    Returns
    -------
        : np.ndarray
            The lattice site positions of the given lattice vectors of shape (M, 2)
    """

    lat = np.zeros_like(ab)

    lat[:, 0] = origin[0] + np.round(ab[:, 0] * fraction[0]) * (u[0] / fraction[0]) + \
                np.round(ab[:, 1] * fraction[1]) * v[0] / fraction[1]
    lat[:, 1] = origin[1] + np.round(ab[:, 0] * fraction[0]) * (u[1] / fraction[0]) + \
                np.round(ab[:, 1] * fraction[1]) * v[1] / fraction[1]

    return lat


def lattice2D_2(u, v, a, b, xy0, numPoints):
    """ A modified version of lattice2D which uses non-normalized u, v vectors

    """

    Y2D, X2D = np.meshgrid(range(numPoints[1]), range(numPoints[0]), indexing=default_indexing)

    X = X2D.ravel()
    Y = Y2D.ravel()

    xy = np.zeros((np.prod(numPoints), 2))
    xy[:, 0] = xy0[0] + a * X * u[0] + b * Y * v[0]
    xy[:, 1] = xy0[1] + a * X * u[1] + b * Y * v[1]

    return xy


def lattice3D_2(u, v, w, a, b, c, xyz0, numPoints):
    """ A modified version of lattice3D which uses non-normalized u,v vectors

    """

    Z3D, Y3D, X3D = np.meshgrid(range(numPoints[1]), range(numPoints[0]), range(numPoints[0]),
                                indexing=default_indexing)

    X = X3D.ravel()
    Y = Y3D.ravel()
    Z = Z3D.ravel()

    xyz = np.zeros((np.prod(numPoints), 3))
    xyz[:, 0] = xyz0[0] + a * X * u[0] + b * Y * v[0] + c * Z * w[0]
    xyz[:, 1] = xyz0[1] + a * X * u[1] + b * Y * v[1] + c * Z * w[1]
    xyz[:, 2] = xyz0[1] + a * X * u[2] + b * Y * v[2] + c * Z * w[2]

    return xyz


def fit_peaks_gauss2d(image, peaks, cutOut, init, bounds, remove_edge_peaks=True):
    """ Fit peaks to a 2D Gaussian. The Gaussian function is gaussND.gauss2D()
    
    Todo: Write tests. This function is copied from test_peakFind.ipynb
    Todo: Add example test_peakFind.ipynb from my jupyter notebook folder
    
    Parameters
    ----------
        image : np.ndarray
            The 2D image with the intensities to fit to
            
        peaks : np.ndarray
            The peaks in a ndarray of shape (N,2) where N is the number
            of peaks. Should match output of peakFind.peakFind2D.
            
        cutOut : int
            The size (+/-) of the region around each peak to fit.
            
        init : tuple
            A tuple of initial sigma values in x and y. (sigma_x, sigma_y)
        
        bounds : tuple
            A tuple of 2 tuples indicating the upper and lower bounds for the fitting.
            See scipy.optimize.curve_fit for more information. Ordered as
            ( (center_x_low, center_y_low, sigma_x_low, sigma_y_low), 
            (center_x_high, center_y_high, sigma_x_high, sigma_y_high))
        
        remove_edge_peaks : bool, default = True
            Peak positions at the edge of the image are set to NaN and are removed.
            If you want to have those positions returned as nan set this to False.
        
    Note
    ----
        This function uses np.meshgrid and indexing='ij' internally.
    
    Returns
    -------
        : tuple
            Returns a tuple of 3 arrays:
                [0] optimized peak positions as a (M, 2) ndarray. M <= N. Peaks on the edge of the image are removed.
                [1] peak intensity value interpolated at the optimized peak position.
                [2] the fitting values for each peak
    """

    # Indexes that sort intensity high to low
    peaks_sort_values = np.argsort(image[peaks[:, 0], peaks[:, 1]])

    # X,Y,sig_x,sig_y
    # NAN means peak near an edge and can not be fit
    optPointsNaN = np.zeros((peaks_sort_values.shape[0], 4))
    fittingValues = np.zeros_like(optPointsNaN)

    optINaN = np.zeros((optPointsNaN.shape[0]))  # holds the optimized intensity
    optI_meanNaN = np.zeros_like(optINaN)

    # Setup cutout coordinates around each peak
    X2D, Y2D = np.meshgrid(np.arange(-cutOut, cutOut + 1, 1),
                           np.arange(-cutOut, cutOut + 1, 1), indexing=default_indexing)

    for ii, index in enumerate(peaks_sort_values):

        curX = int(peaks[index, 0])
        curY = int(peaks[index, 1])

        # Check if current point + fit region is within the bounds of the volume
        if (curX >= cutOut) & (curY >= cutOut) & \
                (curX < image.shape[0] - cutOut) & \
                (curY < image.shape[1] - cutOut):

            curIm = np.float32(image[curX - cutOut:curX + cutOut + 1,
                               curY - cutOut:curY + cutOut + 1])
            curIm_norm = curIm - curIm.min()  # subtract the min
            curIm_norm = curIm_norm / curIm_norm.max()  # normalize to 1

            # Calculate error for each point based on intensity
            y_error = 1 / np.sqrt(curIm_norm.ravel() + 0.00001)  # add a small offset to remove divide by zero
            
            optP2D, optCov2D = opt.curve_fit(gaussND.gauss2D_FIT, (X2D, Y2D), curIm_norm.ravel(),
                                             p0=(0, 0, init[0], init[1]), bounds=bounds,
                                             sigma=y_error)

            fittingValues[ii, :] = optP2D

            optPointsNaN[ii, :] = np.array([curX + optP2D[0], curY + optP2D[1],
                                            optP2D[2], optP2D[3]])  # X,Y,sigma_x,sigma_y
        else:
            # Peak is near an edge
            optPointsNaN[ii, :] = np.array([np.nan] * 4)
            fittingValues[ii, :] = np.array([np.nan] * 4)
            optINaN[ii] = np.nan
            optI_meanNaN[ii] = np.nan

    # Get the intensity of each optimize peak position
    optINaN = ndimage.map_coordinates(image.astype('f'),
                                      ((optPointsNaN[:, 0]),
                                       (optPointsNaN[:, 1])),
                                      cval=np.nan)

    # Remove NANs where peak is at the edge if desired
    if remove_edge_peaks:
        optPoints = optPointsNaN[~np.isnan(optPointsNaN[:, 0]), :]
        optI = optINaN[~np.isnan(optINaN)]
        fittingValues = fittingValues[~np.isnan(optINaN)]
    else:
        optPoints = optPointsNaN
        optI = optINaN
        fittingValues = fittingValues

    return optPoints, optI, fittingValues


def fit_peaks_gauss3d(volume, peaks, cutOut, init, bounds, remove_edge_peaks=True):
    """ Fit peaks to a 2D Gaussian. The Gaussian function is gaussND.gauss2D()

    Todo: Write tests. This function is copied from test_peakFind_3d.ipynb

    Parameters
    ----------
        volume : np.ndarray
            The 3D image with the intensities to fit to.

        peaks : np.ndarray
            The peaks in a ndarray of shape (N, 3) where N is the number
            of peaks. Should match output of peakFind.peakFind3D.

        cutOut : int
            The size (+/-) of the region around each peak to fit.

        init : tuple
            A 3-tuple of initial sigma values in x and y. (sigma_x, sigma_y, sigma_z)

        bounds : tuple
            A 2-tuple of 6-tuples indicating the upper and lower bounds for the fitting.
            See scipy.optimize.curve_fit for more information. Ordered as
            ( (center_x_low, center_y_low, center_z_low, sigma_x_low, sigma_y_low, sigma_z_low),
            (center_x_high, center_y_high, center_z_high, sigma_x_high, sigma_y_high, sigma_z_high))

        remove_edge_peaks : bool, default = True
            Peak positions at the edge of the image are set to NaN and are removed.
            If you want to have those positions returned as nan set this to False.

    Note
    ----
        This function uses np.meshgrid and indexing='ij' internally.

    Returns
    -------
        : tuple
            Returns a tuple of 3 arrays:
                [0] optimized peak positions as a (M, 3) ndarray. M <= N. Peaks on the edge of the image are removed
                unless otherwise specified. Then the edge peaks are returned as NAN values.
                [1] peak intensity value interpolated at the optimized peak position.
                [2] the local fitting values for each peak.
    """
    # Sort from highest to lowest intensity
    validPeaks_sort_values = np.argsort(volume[peaks[:, 0], peaks[:, 1], peaks[:, 2]])

    # X,Y,Z,sig_x,sig_y,sig_z
    # NAN means peak near an edge and can not be fit
    optPointsNaN = np.zeros((len(validPeaks_sort_values), 6))
    fittingValues = np.zeros_like(optPointsNaN)

    # Set up fitting static variables
    local_region = np.arange(-cutOut, cutOut + 1, 1)
    Y3D, X3D, Z3D = np.meshgrid(local_region, local_region, local_region, indexing=default_indexing)
    bad_point = (np.nan, np.nan, np.nan, np.nan, np.nan, np.nan)

    for ii, index in enumerate(validPeaks_sort_values):
        curX = int(peaks[index, 0])
        curY = int(peaks[index, 1])
        curZ = int(peaks[index, 2])

        # Check if current point + fit region is within the bounds of the volume
        if (curX > cutOut + 1) & (curY > cutOut + 1) & (curZ > cutOut + 1) & (curX < volume.shape[0] - cutOut) \
                & (curY < volume.shape[1] - cutOut) & (curZ < volume.shape[2] - cutOut):

            curVol = np.float32(volume[curX - cutOut:curX + cutOut + 1,
                                curY - cutOut:curY + cutOut + 1,
                                curZ - cutOut:curZ + cutOut + 1])
            curVol_norm = curVol - curVol.min()  # subtract the min
            curVol_norm = curVol_norm / curVol_norm.max()  # normalize maximum to 1

            # Fit to a 3D Gaussian to the area around the peak
            y_error = 1 / np.sqrt(curVol_norm.ravel() + 0.00001)  # add error
            optP3D, optCov3D = opt.curve_fit(gaussND.gauss3D_FIT, (X3D, Y3D, Z3D), curVol_norm.ravel(),
                                             p0=(0, 0, 0, *init), bounds=bounds, sigma=y_error)
            fittingValues[ii, :] = optP3D

            # Enter values into output array as X,Y,Z,sigma_x,sigma_y,sigma_z
            optPointsNaN[ii, :] = np.array((float(curX) + optP3D[0], float(curY) + optP3D[1], float(curZ) + optP3D[2],
                                            optP3D[3], optP3D[4], optP3D[5]))

        else:
            # if the peak is near an edge then ignore it and set its values to NAN
            optPointsNaN[ii, :] = np.array(bad_point)
            fittingValues[ii, :] = np.array(bad_point)

    # Get the intensity of each optimize peak position
    optINaN = ndimage.map_coordinates(volume.astype('f'),
                                      ((optPointsNaN[:, 0]),
                                       (optPointsNaN[:, 1]),
                                       (optPointsNaN[:, 2])),
                                      cval=np.nan)

    # Remove NANs where peak is at the edge if desired
    if remove_edge_peaks:
        optPoints = optPointsNaN[~np.isnan(optPointsNaN[:, 0]), :]
        optI = optINaN[~np.isnan(optINaN)]
        fittingValues = fittingValues[~np.isnan(optINaN)]
    else:
        optPoints = optPointsNaN
        optI = optINaN
        fittingValues = fittingValues

    return optPoints, optI, fittingValues


def match_lattice_peaks(peaks, u, v, origin):
    """ Find the matching lattice points to the experimental peaks. This is useful to generate all fitted lattice points
    for a set of experimental peaks. This can then be used directly to calculate displacements for example.

    Parameters
    ----------
    peaks : ndarray
        The experimental peaks as a (num_peaks, 2) ndarray
    u, v : tuple
        The u and v vectors for the fitted lattice
    origin : tuple
        The origin of the fitted lattice for u, v vectors

    Returns
    -------
    : ndarray
        An array of shape (num_peaks, 2) where each coordinate is the closest lattice point for each input peak.
    """

    uv = np.asarray((u, v))
    ab_nearest = np.dot(peaks - origin, np.linalg.inv(uv))

    return ab_nearest


def latticeDisplacements(peaks, u, v, origin):
    """ Find the displacements of the experimental peaks to the fitted lattice parameters

    Parameters
    ----------
    peaks : ndarray
        The experimental peaks with shape (num_peaks, 2). For example, the output of fit_peaks_gauss2d.
    u, v : tuple
        The u and v vectors for the fitted lattice
    origin : tuple
        The origin of the fitted lattice for u, v vectors

    Returns
    -------
    : ndarray
        The displacement of each peak from the expected lattice position.
    """
    ab_nearest = match_lattice_peaks(peaks, u, v, origin)

    uv = np.asarray((u, v))

    p0_disp_linalg = np.zeros(peaks.shape)
    # p0_disp_r_theta_linalg = np.zeros(peaks.shape)
    for ii in range(peaks.shape[0]):
        pp = peaks[ii, :]
        rref = np.dot(np.round(ab_nearest[ii, :]), uv) + origin
        p0_disp_linalg[ii, :] = pp - rref
        # p0_disp_r_theta_linalg[ii, :] = (np.sqrt((pp[0] - rref[0]) ** 2 + (pp[1] - rref[1]) ** 2),
        #                                  np.arctan2((pp[1] - rref[1]), (pp[0] - rref[0])))
    return p0_disp_linalg


def calculate_unit_cell(image, lattice, u, v, unit_cell_size):
    """ IN PROGRESS
    Calculate a unit cell using the input lattice and lattice parameters. Any unit cell at the edge of the image is not
    used as part of the unit cell will be invalid.

    Parameters
    ----------
    image : ndarray
        The image containing the unit cell in a regular pattern (i.e. a lattice)
    lattice: ndarray
        The starting position of each unit cell in the image.
    u, v : tuple
        The u and v vectors for the fitted lattice
    unit_cell_size : int or tuple
        If int then this is the number of pixels in the unit cell along the u vector. The size of the v vector will be
        automatically calculated to make the unit cell pixels square. If tuple then the number of pixels along each
        u and v dimension are used directly.

    Returns
    -------
    : ndarray
        An array of size unit_cell_size which contains the average value of each position in the unit cell for each
        peak position and u,v, lattice point.
    """
    print('Warning, untested code. Work in progress.')
    if isinstance(unit_cell_size, int):
        num_u = unit_cell_size
        num_v = int(num_u * (np.linalg.norm(v) / np.linalg.norm(u)))
    elif isinstance(unit_cell_size, tuple):
        num_u, num_v = unit_cell_size
    else:
        raise TypeError('unit_cell_size must be int or 2-tuple')

    # Ensure image is floating point for interpolation
    if np.issubdtype(image.dtype, np.floating):
        image2 = image.copy()
    else:
        image2 = image.astype(np.float32)

    uu = [ii / num_u for ii in u]
    vv = [ii / num_v for ii in v]

    # Create a centered set of sub-lattice coordinates
    sub_lattice = lattice2D(uu, vv, 1, 1, (0, 0), (num_u, num_v))  # starts at (0, 0). Then offset for each peak.

    unit_cell = np.zeros((num_v, num_u), dtype=image2.dtype)
    cur_cell = np.zeros((num_v * num_u,), dtype=image2.dtype)
    num = 0  # number of unit cells used (edges are not used)
    for ii, peak in enumerate(lattice):
        cur_XX = sub_lattice[:, 0] + peak[0]
        cur_YY = sub_lattice[:, 1] + peak[1]
        ndimage.map_coordinates(image2, (cur_XX.ravel(), cur_YY.ravel()), order=3, output=cur_cell)
        if not np.any(np.isnan(cur_cell)):
            # Avoid unit cells on the edge of the image. Otherwise the mean is corrupted
            unit_cell += cur_cell.reshape((num_v, num_u))
            num += 1

    # Normalize by the number of unit cells used
    unit_cell /= num

    # Rotate to match the image
    return unit_cell.T



"""
Module to calculate radial profiles.

The **settings** dict used for the evaluation has the following entries:

    * lmax_r (int):    Radius used for local maximum detection in [px].
    * lmax_thresh (int):    Threshold for local maximum detection in [intensity].
    * lmax_cinit (int, int):    Initial guess for center in [px].
    * lmax_range (float, float):    r range to cinit used to filter local maxima [dims].
    * ns (int, ...):    Distortion orders to correct.
    * fit_rrange (float, float):    r range used to fit radial profile [dims].
    * back_xs (float, ...):    Fixpoints for fitting background [dims].
    * back_xswidth (float):    Range around fixpoints to take into account [dims].
    * back_init (float):    Initial guess for subtracting background.
    * fit_funcs (str, ...):    List of functions to model radial profile.
    * fit_init (float, ...):    Initial guess for fitting.

    optional (set to None to use defaults):

        * plt_imgminmax (float, float):    Relative range to plot the img.
        * rad_rmax (float):    Max r to be used in radial profile [dims].
        * rad_dr (float): Stepsize for r-axis in radial profile [dims].
        * rad_sigma (float): Sigma for Gaussian used as kernel density estimator [dims].
        * mask (np.ndarray): Binary image as img, 0 for pixels to exclude.
        * fit_maxfev (int): Maxfev forwarded to scipy optimize.

"""

import copy
import numpy as np
import scipy.ndimage
import scipy.interpolate

import ncempy.algo.math
import ncempy.algo.distortion
import ncempy.viz


def calc_polarcoords(center, dims, ns=None, dists=None):
    """Calculate the polar coordinates for an image of given shape.

    Center is assumed to be in real coordinates (if not just fake the dims).
    Distortions are corrected if ns and corresponding dists are given.

    Parameters
    ----------
        center : np.ndarray/tuple
            Center of polar coordinate system.
        dims : tuple
            Tuple of dimensions.
        ns : tuple
            List of distortion orders to correct for.
        dists : np.ndarray
            Parameters for distortions.

    Returns
    --------
        : tuple
            Polar coordinates (r, theta) as two np.ndarrays with same dimensions as original image.

    """

    # check input
    try:
        # check center 
        center = np.array(center)
        center = np.reshape(center, 2)

        # check if enough dims available
        assert (len(dims) >= 2)
        assert (len(dims[0]) == 3)

        # check orders
        if ns is not None:
            assert (len(ns) >= 1)

            # check dists
            assert (dists.shape[0] == len(ns) * 2 + 1)
    except TypeError:
        raise TypeError('Something wrong with the input!')

    # create coordinates
    xx, yy = np.meshgrid(dims[0][0], dims[1][0], indexing='ij')

    # calculate polar coordinate system
    rs = np.sqrt(np.square(xx - center[0]) + np.square(yy - center[1]))
    thes = np.arctan2(yy - center[1], xx - center[0])

    # correct for distortions
    if ns is not None:
        for i in range(len(ns)):
            rs /= ncempy.algo.distortion.rad_dis(thes, dists[i * 2 + 1], dists[i * 2 + 2], ns[i])

    return rs, thes


def correct_distortion(img, dims, center, ns, dists):
    """Give corrected version of diffraction pattern with respect to distortions.

    Parameters
    -----------
        img : np.ndarray
            Image (2D).
        dims : tuple
            Dimensions tuple.
        center : np.ndarray or tuple
            Center to be used.
        ns : tuple
            List of distortion orders.
        dists : np.ndarray
            Distortion parameters.

    Returns
    --------
        : np.ndarray
           Corrected diffraction pattern.

    """

    # check input
    try:
        assert (isinstance(img, np.ndarray))

        # check center 
        center = np.array(center)
        center = np.reshape(center, 2)

        # check if enough dims available
        assert (len(dims) >= 2)
        assert (len(dims[0]) == 3)

        # check orders
        assert (len(ns) >= 1)

        # check dists
        assert (dists.shape[0] == len(ns) * 2 + 1)
    except TypeError:
        raise TypeError('Something wrong with the input!')

    rs, thes = calc_polarcoords(center, dims)

    # anti distort
    for i in range(len(ns)):
        rs *= ncempy.algo.distortion.rad_dis(thes, dists[i * 2 + 1], dists[i * 2 + 2], ns[i])

    dis_xx = rs * np.cos(thes) + center[0]
    dis_yy = rs * np.sin(thes) + center[1]

    f = scipy.interpolate.RectBivariateSpline(dims[0][0], dims[1][0], img)

    return f.ev(dis_xx, dis_yy)


def calc_radialprofile(img, rs, rMax, dr, rsigma, mask=None):
    """Calculate the radial profile using Gaussian kernel density estimator.

    It is suggested to use an rMax such that all directions are still in the image, otherwise outer areas will contribute differently and lead to different signal-to-noise. A value of dr corresponding to 1/10 px and rsigma corresponding to 1 px are good parameters.

    Parameters
    -----------
        img : np.ndarray
            Image to take radial profile of intensity from.
        rs : np.ndarray
            Array containing the radial distance, same shape as img.
        rMax : float
              Maximum radial distance used for the radial profile.
        dr : float
            Step size for r-axis of radial distance.
        rsigma : float
            Sigma for Gaussian used as kernel density estimator.
        mask : np.ndarray
            Mask image, values at 0 are excluded from radial profile.

    Returns
    --------
        : tuple
            Tuple of radial and intensity axes.

    """

    # check input
    try:
        assert (isinstance(img, np.ndarray))
        assert (isinstance(rs, np.ndarray))
        assert (np.array_equal(img.shape, rs.shape))

        rMax = float(rMax)
        dr = float(dr)
        rsigma = float(rsigma)

        if not mask is None:
            assert (isinstance(mask, np.ndarray))
            assert (np.array_equal(img.shape, mask.shape))

    except:
        raise TypeError('Something wrong with input.')

    # process mask
    if mask is None:
        mask = np.ones(img.shape)
    mask = mask.astype('float64')
    mask[np.where(mask > 0)] = 1
    mask[np.where(mask == 0)] = np.nan

    # prepare radial axis for hist  
    rBins = np.arange(0, rMax, dr)

    radialIndex = np.round(rs / dr) + 1
    rBinMax = len(rBins)
    sel = (radialIndex <= rBinMax)

    sel = np.logical_and(sel, np.logical_not(np.isnan(mask)))

    signal = np.histogram(rs[sel], rBins, weights=img[sel])
    count = np.histogram(rs[sel], rBins, weights=np.ones(img[sel].shape))

    signal_sm = scipy.ndimage.gaussian_filter1d(signal[0], rsigma / dr)
    count_sm = scipy.ndimage.gaussian_filter1d(count[0], rsigma / dr)

    # masked regions lead to 0 in count_sm, divide produces nans, just ignore the warning    
    old_err_state = np.seterr(divide='ignore', invalid='ignore')
    signal_sm = np.divide(signal_sm, count_sm)
    np.seterr(**old_err_state)

    return rBins[:-1], signal_sm


def residuals_fit(param, r, intens, funcs):
    """Residual function to fit radial profile with flexibility.

    The arguments for the single functions are automatically cut from the fit parameters.

    Parameters
    -----------
        param : np.ndarray
            Fit parameters.
        r : np.ndarray
            r-axis.
        intens : np.ndarray
            Intensity-axis.
        funcs : tuple
            List of functions to include.

    Returns
    --------
        : np.ndarray
            Residuals.

    """

    return intens - ncempy.algo.math.sum_functions(r, funcs, param)


def fit_radialprofile(r, intens, funcs, init_guess, maxfev=None):
    """Fit the radial profile.

    Convenience wrapper for fitting.

    Parameters
    -----------
        r : np.ndarray
            r-axis of radial profile.
        intens : np.ndarray
            Intensity-axis of radial profile.
        funcs : tuple
            List of functions.
        init_guess : np.ndarray
            Initial guess for parameters of functions in funcs.
        maxfev : int
           maximum function calls (for scipy.optimize.leastsq)

    Returns
    --------
        : np.ndarray
            Optimized parameters.

    """

    try:
        # check data
        assert (isinstance(r, np.ndarray))
        assert (isinstance(intens, np.ndarray))
        assert (np.array_equal(r.shape, intens.shape))

        # funcs and params
        assert (len(funcs) >= 1)
        for i in range(len(funcs)):
            assert (funcs[i] in ncempy.algo.math.lkp_funcs)

        init_guess = np.array(init_guess)
        init_guess = np.reshape(init_guess, sum(map(lambda x: ncempy.algo.math.lkp_funcs[x][1], funcs)))

    except:
        raise TypeError('Something wrong with the input!')

    if maxfev is None:
        maxfev = 1000

    popt, flag = scipy.optimize.leastsq(residuals_fit, init_guess, args=(r, intens, funcs), maxfev=maxfev)

    if flag not in [1, 2, 3, 4]:
        print('WARNING: fitting of radial profile failed.')

    return popt


def run_singleImage(img, dims, settings, show=False):
    """Evaluate a single ring diffraction pattern with given settings.

    Parameters
    -----------
        img : np.ndarray
            Image.
        dims : tuple
            Corresponding dim vectors.
        settings : dict
            Dict of settings necessary for the evaluation.
        show : bool
            Set to directly show plots interactively.

    Returns
    --------
        (np.ndarray):    Optimized parameters of fitting the radial profile according to the settings.

    """

    try:
        assert (isinstance(img, np.ndarray))

        # check if dims availabel
        assert (len(dims) >= 1)
        assert (len(dims[0]) == 3)

        assert (type(settings) is dict)

    except:
        raise RuntimeError('Something wrong with the input')

    # create a copy of settings to decouple
    mysettings = copy.deepcopy(settings)

    # get local maxima an turn them into real space coords
    points = ncempy.algo.local_max.local_max(img, mysettings['lmax_r'], mysettings['lmax_thresh'])
    points = ncempy.algo.local_max.points_todim(points, dims)

    # convert center to real space
    center_init = ncempy.algo.local_max.points_todim(mysettings['lmax_cinit'], dims)

    # filter to single ring
    points = ncempy.algo.distortion.filter_ring(points, center_init, mysettings['lmax_range'])

    if show:
        if settings['plt_imgminmax'] is None:
            mysettings['plt_imgminmax'] = (0., 1.)

        plot = ncempy.viz.plot_points(img, points, vminmax=mysettings['plt_imgminmax'], dims=dims, invert=True,
                                      show=show)

    # optimize center
    center = ncempy.algo.distortion.optimize_center(points, center_init, verbose=show)

    # fit distortions
    points_plr = ncempy.algo.distortion.points_topolar(points, center)
    dists = ncempy.algo.distortion.optimize_distortion(points_plr, mysettings['ns'])
    if show:
        plot = ncempy.viz.plot_distpolar(points_plr, dims, dists, mysettings['ns'], show=show)

    # calc coordinates in optimized system
    rs, thes = ncempy.algo.radial_profile.calc_polarcoords(center, dims, mysettings['ns'], dists)

    if settings['rad_rmax'] is None:
        mysettings['rad_rmax'] = np.abs(dims[0][0][0] - dims[0][0][1]) * np.min(img.shape) / 2.0
    if settings['rad_dr'] is None:
        mysettings['rad_dr'] = np.abs(dims[0][0][0] - dims[0][0][1]) / 10.
    if settings['rad_sigma'] is None:
        mysettings['rad_sigma'] = np.abs(dims[0][0][0] - dims[0][0][1])

    # extract radial profile
    R, I = ncempy.algo.radial_profile.calc_radialprofile(img, rs, mysettings['rad_rmax'], mysettings['rad_dr'],
                                                         mysettings['rad_sigma'], mask=mysettings['mask'])

    # cut radial profile
    sel = (R >= mysettings['fit_rrange'][0]) * (R <= mysettings['fit_rrange'][1])
    I = I[sel]
    R = R[sel]

    rawRI = np.copy(np.array([R, I]).transpose())

    # subtract a power law background fitted to specific points
    fit_R = np.array([])
    fit_I = np.array([])
    for xpoint in mysettings['back_xs']:
        ix = np.where(np.abs(R - xpoint) < mysettings['back_xswidth'])
        fit_R = np.append(fit_R, R[ix])
        fit_I = np.append(fit_I, I[ix])

    funcs_back = ['const', 'powlaw']
    res_back = ncempy.algo.radial_profile.fit_radialprofile(fit_R, fit_I, funcs_back, mysettings['back_init'],
                                                            maxfev=1000)
    if show:
        plot = ncempy.viz.plot_fit(R, I, dims, funcs_back, res_back, show=show)

    I = I - ncempy.algo.math.sum_functions(R, funcs_back, res_back)

    # fit
    res = ncempy.algo.radial_profile.fit_radialprofile(R, I, mysettings['fit_funcs'], mysettings['fit_init'],
                                                       maxfev=mysettings['fit_maxfev'])

    if show:
        plot = ncempy.viz.plot_fit(R, I, dims, mysettings['fit_funcs'], res, show=show)

    return np.array([R, I]).transpose(), res, center, dists, rawRI, res_back, mysettings



""" A module with Gaussian functions for creating Gaussian distributions
in 1D, 2D and 3D. There are also functions with the same name and _FIT
which can be used to fit an intensity distribution in a ndarray with the
corresponding Gaussian distribution using scipy's curve_fit.

A 1D Lorentz and a 1D gaussLorentz function is also included for fitting
zero loss peaks in EELS spectra.

These functions were written assuming you use np.meshgrid() with indexing = 'xy'. If
you use 'ij' indexing then be careful about how you pass in the x and y coordinates.

"""

import numpy as np


def gauss1D(x, x0, sigma):
    """ Returns the value of a gaussian at a 2D set of points for the given standard deviation with maximum
    normalized to 1.
     
    Parameters
    ----------
        x: ndarray
            A vector of size (N,) of points for the x values
        x0: float
            The center of the Gaussian.
        sigma: float
            Standard deviation of the Gaussian.
            
    Returns
    -------
        g: ndarray
            A vector of size (N,) of the Gaussian distribution evaluated
            at the input x values.
            
    Note
    ----
        Calculate the half width at half maximum (HWHM) as
        >> HWHM = sqrt(2*log(2))*stDev ~ 1.18*stDev
        or if x goes from -1 to 1
        >> 0.5*size(x)*stDev
        
    """
    return np.exp(-(x - x0) ** 2 / (2 * sigma ** 2))


def lorentz1D(x, x0, w):
    """ Returns the probability density function of the Lorentzian (aka Cauchy) function with maximum at 1.
    
    Parameters
    ----------
        x: ndarray or list
            A 1D vector of points for the dependent variable
        x0: float
            The center of the peak maximum
        w: float
            The parameter that modifies the width of the distribution.
            
    Returns
    -------
        l: ndarray
            A vector of size (N,) of the Lorentzian distribution evaluated 
            at the input x values.
            
    """
    return w ** 2 / ((x - x0) ** 2 + w ** 2)


def gaussLorentz1D(x, x0, w):
    """ A Gaussian-Lorentzian function in one dimension.

    Parameters
    ----------
        x: ndarray or list
            A 1D vector of points for the dependent variable
        x0: float
            The center of the peak maximum
        w: float
            The parameter that modifies the width of the distribution.

    Returns
    -------
        lg: ndarray
            A vector of size (N,) of the Lorentzian Gaussian distribution
            evaluated at the input x values.

    """
    gSigma = w / (2 * np.sqrt(2 * np.log(2)))  # change the width to sigma
    return gauss1D(x, x0, gSigma) + lorentz1D(x, x0, w)


def gauss2D(x, y, x0, y0, sigma_x, sigma_y):
    """ Calculates the value of a Gaussian at a 2D set of points for the given
    standard deviations with maximum normalized to 1.
    The Gaussian axes are assumed to be 90 degrees from each other.
    
    Parameters
    ----------
        x, y : ndarray
            A 2D array of size (M,N) of points for the x values. 
            Use x, y = np.meshgrid(range(M),range(N),indexing='xy').
        x0 : float
            The center of the Gaussian along the x direction.
        y0 : float
            The center of the Gaussian along the y direction.
        sigma_x : float
            Standard deviation of the Gaussian along x.
        sigma_y : float
            Standard deviation of the Gaussian along y.
            
    Returns
    -------
        g : ndarray
            A ndarray of size (N, M) of the Gaussian distribution evaluated
            at the input x values.
            
    Note
    ----
        The Gaussian is normalized such that the peak == 1. To normalize the
        integral divide by 2 * np.pi * sigma_x * sigma_y
        
    """
    x0 = float(x0)
    y0 = float(y0)
    g2 = np.exp(-((x - x0) ** 2 / (2 * sigma_x ** 2) + (y - y0) ** 2 / (2 * sigma_y ** 2)))
    g2_norm = g2 / np.max(g2.flatten())
    return g2_norm


def gauss2D_FIT(xy, x0, y0, sigma_x, sigma_y):
    """ Version of gauss2D used for fitting (1 x_data input (xy)
    and flattened output). Returns the value of a gaussian at a 2D set of points for the given
    standard deviation with maximum normalized to 1.
    The Gaussian axes are assumed to be 90 degrees from each other.
    
    Parameters
    ----------
        xy: tuple
            A (N,M) shaped tuple containing the vectors of points for the
            evaluation points.
        x0: float
            The x center of the Gaussian.
        y0: float
            The y center of the Gaussian.
        sigma_x: float
            The standard deviation of the Gaussian along x.
        sigma_y: float
            The standard deviation of the Gaussian along y.
    
    Returns
    -------
        g2_norm: ndarray
            The Gaussian distribution with maximum value normalized to 1. The
            2D ndarray is reshaped into a (N*M,) array for use in fitting
            functions in numpy and scipy.
        
    """
    x0 = float(x0)
    y0 = float(y0)
    x = xy[0]
    y = xy[1]
    g2 = np.exp(-((x - x0) ** 2 / (2 * sigma_x ** 2) + (y - y0) ** 2 / (2 * sigma_y ** 2)))
    g2_norm = g2 / np.max(g2.flatten())
    return g2_norm.reshape(-1)


def gauss2D_theta(x, y, x0, y0, sigma_x, sigma_y, theta):
    """ Returns the value of a generalized Gaussian at a 2D set of points for
    the given standard deviation with maximum normalized to 1.
    The Gaussian axes (assumed to be 90 degrees from each other) can be
    oriented at different angles to the output array axes using theta.
    
    Parameters
    ----------
        x: ndarray or list
            Evaluation points along x of size (N,)
        y: ndarray
            Evaluation points along y of size (M,)
        x0: float
            Center of the maximum along x.
        y0: float
            Center of the maximum along y.
        sigma_x: float
            Standard deviation along x.
        sigma_y: float
            Standard deviation along y.
        theta: float
            The rotation of the Gaussian principle axes from the array
            horizontal and vertical. In radians.
            
    Returns
    -------
        g2_norm: ndarray
            A (N, M) sized 2D ndarray with a Gaussian distribution rotated.
            
    """

    x0 = float(x0)
    y0 = float(y0)
    a = (np.cos(theta) ** 2) / (2 * sigma_x ** 2) + (np.sin(theta) ** 2) / (2 * sigma_y ** 2)
    b = -(np.sin(2 * theta)) / (4 * sigma_x ** 2) + (np.sin(2 * theta)) / (4 * sigma_y ** 2)
    c = (np.sin(theta) ** 2) / (2 * sigma_x ** 2) + (np.cos(theta) ** 2) / (2 * sigma_y ** 2)
    g2 = np.exp(- (a * ((x - x0) ** 2) + 2 * b * (x - x0) * (y - y0) + c * ((y - y0) ** 2)))
    g2_norm = g2 / np.max(g2.flatten())
    return g2_norm  # return a 2D array


def gauss2D_theta_FIT(xy, x0, y0, sigma_x, sigma_y, theta):
    """ Version of gauss2D_theta used for fitting (1 x_data input and
    flattened output). Returns the value of a gaussian at a 2D set of points for the given standard deviation with
    maximum normalized to 1. The Gaussian axes can be oriented at different angles (theta) in radians.
    
    Parameters
    ----------
        xy: tuple
            Evaluation points along x and y of shape (N,M)
        x0: float
            Center of the maximum along x.
        y0: float
            Center of the maximum along y.
        sigma_x: float
            Standard deviation along x.
        sigma_y: float
            Standard deviation along y.
        theta: float
            The rotation of the Gaussian principle axes from the array
            horizontal and vertical. In radians.
            
    Returns
    -------
        g2_norm: ndarray
            A (N, M) sized 2D ndarray with a Gaussian distribution rotated.

    """
    x0 = float(x0)
    y0 = float(y0)
    x = xy[0]
    y = xy[1]
    a = (np.cos(theta) ** 2) / (2 * sigma_x ** 2) + (np.sin(theta) ** 2) / (2 * sigma_y ** 2)
    b = -(np.sin(2 * theta)) / (4 * sigma_x ** 2) + (np.sin(2 * theta)) / (4 * sigma_y ** 2)
    c = (np.sin(theta) ** 2) / (2 * sigma_x ** 2) + (np.cos(theta) ** 2) / (2 * sigma_y ** 2)
    g2 = np.exp(- (a * ((x - x0) ** 2) + 2 * b * (x - x0) * (y - y0) + c * ((y - y0) ** 2)))
    g2_norm = g2 / np.max(g2.flatten())
    return g2_norm.ravel()  # return a 1D vector


def gauss2D_poly_FIT(xy, x0, y0, A, B, C):
    """ Returns the flattened values of an elliptical gaussian at a 2D set of
    points for the given polynomial pre-factors with maximum normalized to 1.
    The Gaussian axes are assumed to be 90 degrees from each other.
    
    The matrix looks like [[A,B],[B,C]].
    See https://en.wikipedia.org/wiki/Gaussian_function
    
    Parameters
    ----------
        xy: tuple
            Evaluation points along x and y of shape (N,M)
        x0: float
            Center of the maximum along x.
        y0: float
            Center of the maximum along y.
        A: float
            The A pre-factor for the polynomial expansion in the exponent.
        B: float
            The B pre-factor for the polynomial expansion in the exponent.
        C: float
            The C pre-factor for the polynomial expansion in the exponent.
            
    Returns
    -------
        g2_norm: ndarray
            A (N, M) sized 2D ndarray with a Gaussaian distribution rotated.
    """

    x0 = float(x0)
    y0 = float(y0)
    x = xy[0]  # retrieve the array from the tuple
    y = xy[1]
    g2 = np.exp(-(A * (x - x0) ** 2 + 2 * B * (x - x0) * (y - y0) + C * (y - y0) ** 2))
    g2_norm = g2 / np.max(g2.flatten())
    return g2_norm.ravel()


def gauss3D(x, y, z, x0, y0, z0, sigma_x, sigma_y, sigma_z):
    """
    gauss3D(x,y,z,x0,y0,z0,sigma_x,sigma_y,sigma_z)
    Returns the value of a gaussian at a 2D set of points for the given
    standard deviations with maximum normalized to 1.
    The Gaussian axes are assumed to be 90 degrees from each other.

    Note
    -----
    Be careful about the indexing used in meshgrid and the order in which you pass the x, y, z variables in.

    Parameters
    ----------
        x, y, z: ndarray, from numpy.meshgrid
            2D arrays of points (from meshgrid)
     
        x0, y0, z0: float
            The x, y, z centers of the Gaussian
        
        sigma_x, sigma_y, sigma_z: float
            The standard deviations of the Gaussian.
    Returns
    -------
        g3_norm: ndarray
            A 3D ndarray
            
    """
    x0 = float(x0)
    y0 = float(y0)
    z0 = float(z0)
    g3 = np.exp(-((x - x0) ** 2 / (2 * sigma_x ** 2) + (y - y0) ** 2 / (2 * sigma_y ** 2) + (z - z0) ** 2 / (
                2 * sigma_z ** 2)))
    g3_norm = g3 / np.max(g3.flatten())
    return g3_norm


def gauss3D_FIT(xyz, x0, y0, z0, sigma_x, sigma_y, sigma_z):
    """
    gauss3D_FIT((x,y,z),x0,y0,z0,sigma_x,sigma_y,sigma_z)
    Returns the value of a gaussian at a 2D set of points for the given
    standard deviations with maximum normalized to 1.
    The Gaussian axes are assumed to be 90 degrees from each other.
     xyz - 
     x0, y0, z0 = the x, y, z centers of the Gaussian
     sigma_x, sigma_y, sigma_z = The std. deviations of the Gaussian.

    Note
    -----
    Be careful about the indexing used in meshgrid and the order in which you pass the x, y, z variables in.

    Parameters
    ----------
        xyz: tuple of ndarrays
            A tuple containing the 3D arrays of points (from meshgrid)
        x0, y0, z0: float
            The x, y, z centers of the Gaussian
        sigma_x, sigma_y, sigma_z: float
            The standard deviations of the Gaussian.
    Returns
    -------
        g3_norm: ndarray
            A flattened array for fitting.
    
    """

    x0 = float(x0)
    y0 = float(y0)
    z0 = float(z0)
    x = xyz[0]
    y = xyz[1]
    z = xyz[2]
    g3 = np.exp(-((x - x0) ** 2 / (2 * sigma_x ** 2) + (y - y0) ** 2 / (2 * sigma_y ** 2) + (z - z0) ** 2 / (
                2 * sigma_z ** 2)))
    g3_norm = g3 / np.max(g3.flatten())
    return g3_norm.ravel()


def gauss3D_poly(x, y, z, x0, y0, z0, A, B, C, D, E, F):
    """
    gauss3Dpoly_FIT((x,y,z),x0,y0,z0,A,B,C,D,E,F)
    Returns the value of a gaussian at a 2D set of points for the given
    standard deviations with maximum normalized to 1.
    The Gaussian axes are not locked to be 90 degrees.

    Parameters
    ----------
        x, y, z : ndarray, 3D
            3D arrays of points (from meshgrid)

        x0, y0, z0 : float
            The x, y, z centers of the Gaussian

        A, B, C, D, E, F : float
            The polynomial values for the fit

    Returns
    -------
        : ndarray, 3D
            The 3D Gaussian
    """
    x0 = float(x0)
    y0 = float(y0)
    z0 = float(z0)
    g3 = np.exp(-(A * (x - x0) ** 2 + B * (y - y0) ** 2 + C * (z - z0) ** 2 + 2 * D * (x - x0) * (y - y0) + 2 * E * (
                x - x0) * (z - z0) + 2 * F * (y - y0) * (z - z0)))
    g3_norm = g3 / np.max(g3.flatten())
    return g3_norm


def gauss3D_poly_FIT(xyz, x0, y0, z0, A, B, C, D, E, F):
    """
    gauss3Dpoly_FIT((x,y,z),x0,y0,z0,A,B,C,D,E,F)
    Returns the value of a gaussian at a 2D set of points for the given
    standard deviations with maximum normalized to 1.
    The Guassian axes are not locked to be 90 degrees.

    Parameters
    ----------
        xyz : tuple of ndarrays
            3D arrays of points (from meshgrid) combined in a tuple

        x0, y0, z0 : float
            The x, y, z centers of the Gaussian

        A, B, C, D, E, F : float
            The polynomial values for the fit

    Returns
    -------
        : ndarray, 3D
            The 3D Gaussian
    """
    x0 = float(x0)
    y0 = float(y0)
    z0 = float(z0)
    x = xyz[0]
    y = xyz[1]
    z = xyz[2]
    g3 = np.exp(-(A * (x - x0) ** 2 + B * (y - y0) ** 2 + C * (z - z0) ** 2 + 2 * D * (x - x0) * (y - y0) + 2 * E * (
                x - x0) * (z - z0) + 2 * F * (y - y0) * (z - z0)))
    g3_norm = g3 / np.max(g3.flatten())
    return g3_norm.ravel()


def gauss3DGEN_FIT(xyz, x0, y0, z0, sigma_x, sigma_y, sigma_z, Angle1, Angle2, Angle3, BG, Height):
    """
    gauss3DGEN_FIT((x,y,z),x0,y0,z0,sigma_x,sigma_y,sigma_z,Angle1,Angle2,Angle3,BG,Height)
    Returns the value of a gaussian at a 3D set of points for the given
    sub-pixel positions with standard deviations, 3D Eular rotation angles,
    constant Background value, and Gaussian peak height.

    adapted from code by Yongsoo Yang, yongsoo.ysyang@gmail.com

    Note
    ----
        This is a work in progress. Needs more testing.

    Parameters
    ----------
        xyz : tuple of 3 np.ndarray
            3D arrays of points (from meshgrid) combined in a tuple

        x0, y0, z0 : float
            The x, y, z centers of the Gaussian
        sigma_x,sigma_y,sigma_z: float
            standard deviations along x,y,z direction before 3D angular rotation
        Angle1, Angle2, Angle3 : float, degrees
            Tait-Bryan angles in ZYX convention for 3D rotation in degrees
        BG : float
            Background
        Height: float
            The peak height of the Gaussian function

    Returns
    -------
        : ndarray, 3D
            The 3D Gaussian
    """
    # 3D vectors for each sampled positions
    print('Warning: this function is using Fortran ordered array for use in tomviz. Need to test this')
    # Todo: Remove Fortran ordering
    v = np.array([xyz[0].reshape(-1, order='F') - x0,
                  xyz[1].reshape(-1, order='F') - y0,
                  xyz[2].reshape(-1, order='F') - z0])

    # rotation axes for Tait-Bryan angles
    vector1 = np.array([0, 0, 1])
    rotmat1 = MatrixQuaternionRot(vector1, Angle1)

    vector2 = np.array([0, 1, 0])
    rotmat2 = MatrixQuaternionRot(vector2, Angle2)

    vector3 = np.array([1, 0, 0])
    rotmat3 = MatrixQuaternionRot(vector3, Angle3)

    # full rotation matrix
    # Todo : Remove dependency on np.matrix()
    rotMAT = np.matrix(rotmat3) * np.matrix(rotmat2) * np.matrix(rotmat1)

    # 3x3 matrix for applying sigmas
    D = np.matrix(np.array([[1. / (2 * sigma_x ** 2), 0, 0, ],
                            [0, 1. / (2 * sigma_y ** 2), 0],
                            [0, 0, 1. / (2 * sigma_z ** 2)]]))

    # apply 3D rotation to the sigma matrix
    WidthMat = np.transpose(rotMAT) * D * rotMAT

    # calculate 3D Gaussian
    RHS_calc = WidthMat * np.matrix(v)
    Result = Height * np.exp(-1 * np.sum(v * RHS_calc.A, axis=0)) + BG

    return Result


def MatrixQuaternionRot(vector, theta):
    """
    Quaternion tp rotate a given theta angle around the given vector.

    adapted from code by Yongsoo Yang, yongsoo.ysyang@gmail.com

    Parameters
    ----------
        vector : ndarray, 3-element
            A non-zero 3-element numpy array representing rotation axis
        theta : float, degrees
            Rotation angle in degrees

    Returns
    -------

     Author: Yongsoo Yang, Dept. of Physics and Astronomy, UCLA
             yongsoo.ysyang@gmail.com
    """
    theta = theta * np.pi / 180
    vector = vector / np.float(np.sqrt(np.dot(vector, vector)))
    w = np.cos(theta / 2)
    x = -np.sin(theta / 2) * vector[0]
    y = -np.sin(theta / 2) * vector[1]
    z = -np.sin(theta / 2) * vector[2]
    RotM = np.array([[1. - 2 * y ** 2. - 2 * z ** 2, 2. * x * y + 2 * w * z, 2. * x * z - 2. * w * y],
                     [2. * x * y - 2. * w * z, 1. - 2. * x ** 2 - 2. * z ** 2, 2. * y * z + 2. * w * x],
                     [2 * x * z + 2 * w * y, 2 * y * z - 2. * w * x, 1 - 2. * x ** 2 - 2. * y ** 2]])

    return RotM
